/* 	$Id: rc.c,v 5.1 2002/09/12 19:53:56 lirc Exp $	 */

/*

rc -  application for sending IR-codes via lirc

Copyright (C) 1998 Christoph Bartelmus (lirc@bartelmus.de)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/



#include <vdr/tools.h>

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <syslog.h>

#include "lirc.h"

#define PACKET_SIZE 256
/* three seconds */
#define TIMEOUT 3


cLirc::cLirc()
: m_szDevice(NULL)
, m_fd(-1)
{

}

cLirc::~cLirc()
{
  close();
  if(m_szDevice) {
	    free(m_szDevice);
	    m_szDevice = NULL;
	}
}

int cLirc::m_nTimeout = 0;

void cLirc::sigalrm(int sig)
{
  m_nTimeout = 1;
}

const char *cLirc::read_string()
{
  static char buffer[PACKET_SIZE + 1] = "";
  char *end;
  static int ptr = 0;
  ssize_t ret;

  if(-1 == m_fd) {
    reopen();
  }

  if(ptr > 0) {
	    memmove(buffer, buffer + ptr, strlen(buffer + ptr) + 1);
	    ptr = strlen(buffer);
	    end = strchr(buffer, '\n');
  } else {
	    end = NULL;
	}
  alarm(TIMEOUT);
  while(end == NULL)
  {
    if(PACKET_SIZE <= ptr) {
		    dsyslog("extb/lirc_read_string(%d): bad packet", m_fd );
		    ptr = 0;
		    return (NULL);
		}
    
    ret = read(m_fd, buffer + ptr, PACKET_SIZE - ptr);

	  if(ret <= 0 || m_nTimeout) {
		  if(m_nTimeout) {
			    dsyslog("extb/lirc_read_string(%d): timeout", m_fd);
		  } else {
			    alarm(0);
			}
		  ptr = 0;
		  return (NULL);
    }
	  buffer[ptr + ret] = 0;
	  ptr = strlen(buffer);
	  end = strchr(buffer, '\n');
  }
  alarm(0);
  m_nTimeout = 0;

  end[0] = 0;
  ptr = strlen(buffer) + 1;

  dsyslog("extb/lirc_read_string(%d): result='%s'", m_fd, buffer);

  return (buffer);
}

enum packet_state {
    P_BEGIN,
    P_MESSAGE,
    P_STATUS,
    P_DATA,
    P_N,
    P_DATA_N,
    P_END
};

int cLirc::send_packet(const char *packet)
{
  int done, todo;
  const char *string, *data;
  char *endptr;
  enum packet_state state;
  int status;
  unsigned long n;
  unsigned long data_n = 0;

  if(-1 == m_fd) {
	    reopen();
	}

  dsyslog("lirc_send_packet(%d): %s", m_fd, packet);

  todo = strlen(packet);
  data = packet;
  while(todo > 0) {
    done = write(m_fd, (void *)data, todo);
	  if(done < 0) {
		  dsyslog("extb/lirc_send_packet(%d): could not send packet: %s",
			  m_fd, strerror(errno));
		  return (-1);
		}
    data += done;
    todo -= done;
	}

  /* get response */
  status = 0;
  state = P_BEGIN;
  n = 0;
  while(1) {
    string = read_string();
    dsyslog("extb/lirc_send_packet/read(%d): %s", m_fd, string);
    
    if(string == NULL)
      return (-1);
    
    switch (state) {
    case P_BEGIN:
      if(strcasecmp(string, "BEGIN") != 0) {
			    continue;
			}
      state = P_MESSAGE;
      break;
		case P_MESSAGE:
      if(strncasecmp(string, packet, strlen(string)) != 0 ||
         strlen(string) + 1 != strlen(packet)) {
			    state = P_BEGIN;
			    continue;
			}
      state = P_STATUS;
      break;
		case P_STATUS:
      if(strcasecmp(string, "SUCCESS") == 0) {
			    status = 0;
		  } else if(strcasecmp(string, "END") == 0) {
        status = 0;
        return (status);
      } else if(strcasecmp(string, "ERROR") == 0) {
        dsyslog("extb/lirc_send_packet(%d): command failed: %s",
        m_fd, packet);
        status = -1;
      } else {
			    goto bad_packet;
			}
      state = P_DATA;
      break;
		case P_DATA:
      if(strcasecmp(string, "END") == 0) {
			    return (status);
      } else if(strcasecmp(string, "DATA") == 0) {
			    state = P_N;
			    break;
			}
      goto bad_packet;
		case P_N:
      errno = 0;
      data_n = strtoul(string, &endptr, 0);
      if(!*string || *endptr) {
			    goto bad_packet;
			}
		  if(data_n == 0) {
        state = P_END;
      } else {
			    state = P_DATA_N;
			}
      break;
		case P_DATA_N:
      dsyslog("extb/lirc_send_packet(%d): %s", m_fd, string);
      n++;
      if(n == data_n)
			state = P_END;
      break;
		case P_END:
      if(strcasecmp(string, "END") == 0) {
			    return (status);
			}
      goto bad_packet;
      break;
		}
	}
bad_packet:
    dsyslog("extb/lirc_send_packet(%d): bad return packet", m_fd);
    return (-1);
}

bool cLirc::open(const char *szDevice)
{
  if(m_szDevice)
	    free(m_szDevice);
  m_szDevice = strdup(szDevice);
  
  struct sockaddr_un addr;
  
  dsyslog("extb/lirc_open");
  
  struct sigaction act;
  act.sa_handler = sigalrm;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;		/* we need EINTR */
  sigaction(SIGALRM, &act, NULL);
  
  addr.sun_family = AF_UNIX;
  strncpy(addr.sun_path, m_szDevice, sizeof(addr.sun_path));
  m_fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if(-1 == m_fd)
  {
    esyslog("extb/lirc_open(%d): could not open socket: %s", m_fd,
    strerror(errno));
    return (false);
  }

  if(connect(m_fd, (struct sockaddr *)&addr, sizeof(addr)) == -1)
  {
    esyslog("extb/lirc_open(%d): could not connect to socket: %s", m_fd,
    strerror(errno));
    return (false);
  }
  return (true);
}

void cLirc::close()
{
  if(-1 == m_fd) {
	    dsyslog("extb/lirc_close(%d)", m_fd);
	    ::close(m_fd);
	    m_fd = -1;
	}
}

void cLirc::send(const char *cmd)
{
    char tmpbuff[256];

    snprintf(tmpbuff, sizeof(tmpbuff), "send_once EXTB_TX %s\n", cmd);
    if(-1 == send_packet(tmpbuff)) {
	    close();
	    reopen();
	  }
}
